Require Import  ArithRing Omega  Max F_omega.

(** About Ackermaan *)

Definition Fsuc' f p := iterate f (S p) 1.

Fixpoint ackermann (n:nat) :=
  match n with
      0 => S
    | S n' => Fsuc' (ackermann n')
  end.

Lemma ack0_0 : forall p, ackermann 0 p = S p.
Proof.   reflexivity. Qed.

Lemma ackS_0 : forall n ,  ackermann (S n) 0 = ackermann n 1.
  intros.
  reflexivity.
Qed .


Lemma ack_S_S : forall n p, ackermann (S n) (S p) =
                            ackermann n (ackermann (S n) p).
Proof .
  intros. reflexivity. 
Qed.
SearchAbout Fsuc. 

Lemma Ack_S_rw : forall i p, ackermann (S i) p= iterate  (ackermann i) (S p) 1.
Proof. 
  reflexivity. 
Qed. 

Lemma Ack_S_rw' : forall i p, ackermann (S i) p= iterate  (ackermann i) p
                                                          (ackermann i 1).
Proof. 
 intros; rewrite Ack_S_rw; now rewrite  iterate_S_eqn2.  
Qed. 

Section On_Fsuc'.
Variable f : nat_fun.
Hypotheses (Hmono: strict_mono f)
           (Hgt :  S <<= f ).

Lemma Fsuc'_mono : strict_mono (Fsuc' f).
Proof.
  intros n p H ; unfold Fsuc'.
   repeat rewrite iterate_S_eqn.
   apply Hmono.
   induction H.
   rewrite iterate_S_eqn.
   apply Hgt. 
   rewrite iterate_S_eqn.
   transitivity (iterate f m 1);auto. 
Qed. 


Lemma Fsuc'_ge_S : S  <<= Fsuc' f.
Proof. 
  intro n; unfold Fsuc'.  rewrite iterate_S_eqn.
  induction n. 
  cbn. 
  specialize (Hgt 0). 
  now apply mono_le. 
  rewrite iterate_S_eqn.
  transitivity (S (f (iterate f n 1))).     
  auto with arith. 
  
  apply Hgt. 
Qed. 

Lemma Fsuc'_le : f <<= Fsuc' f.
Proof. 
  unfold Fsuc'; intro n.
  rewrite iterate_S_eqn.
   induction n.
    cbn. 
     apply Nat.lt_le_incl; apply Hmono. 
     auto with arith. 
     rewrite iterate_S_eqn.
     apply mono_weak; auto. 
     transitivity (f n); auto. 
Qed.


Lemma Fsuc'_dom : dominates_from 1 (Fsuc' f) f.
Proof. 
 unfold Fsuc'; intros n H. rewrite iterate_S_eqn. 
  apply Hmono. 
  induction H.   
   rewrite iterate_S_eqn. 
    cbn. 
    auto. 
    rewrite iterate_S_eqn.
    apply le_lt_trans with (f m); auto with arith. 
Qed. 

End On_Fsuc'.

(* Prove simultaneaously S <= ackermann i and strict_mono (ackermann i) *)

Lemma Ack_gt : forall i, S <<= ackermann i /\ strict_mono  (ackermann i).
induction i. 
 - split.
   + intro n; cbn.
     auto with arith. 
   +   cbn. 
   intros n p H; auto with arith. 
  - 
     destruct IHi; split. 
   + 
     intro n; transitivity (ackermann i n). 
   auto. 
    
   rewrite Ack_S_rw. 
   unfold Fsuc'. 
   rewrite iterate_S_eqn.
   SearchAbout strict_mono.
   apply mono_weak; auto.
   SearchAbout iterate.
   apply R22;auto. 
   + intros x y Hxy. 
      repeat rewrite Ack_S_rw.     
      unfold Fsuc'. 
      SearchAbout (iterate ?f _ _ < iterate ?f _ _).
      apply iterate_lt;auto. 
      auto with arith. 
Qed. 

Lemma Ack_le_F_ : forall i j , 1 <= j -> ackermann i j <= f_ i j.
induction i. 
- cbn; intro i; auto. 
-  cbn . unfold Fsuc', Fsuc . intros. 
   cbn.
   transitivity (f_ i (iterate (ackermann i) j 1)).
   apply IHi.
    clear IHi;
    destruct  i.
    simpl.
    SearchAbout (?i <= iterate _ _ ?i).
   apply iterate_ge.
    intro; auto. 
        apply iterate_ge.
     destruct (Ack_gt (S i));auto. 
   apply f_i_weak_mono.
   transitivity (iterate (f_ i) j 1).
    
   apply iterate_mono_1 with 1; auto.
 destruct (Ack_gt (S i));auto. 
destruct (Ack_gt i);auto. 
destruct (Ack_gt i);auto. 

SearchAbout [strict_mono iterate].
destruct (Lt.le_lt_or_eq _ _ H).
  apply Nat.lt_le_incl. 
   apply iterate_mono. 
destruct (Ack_gt i);auto. 
destruct (Ack_gt i);auto. 
get_properties i.
 auto. 
get_properties i.
 auto. 
 auto. 
 subst. 
 auto with arith. 
Qed. 


Lemma L44 :  forall n, ackermann (S n) 1 = ackermann n (ackermann n 1).
intro; cbn. reflexivity.
Qed.

SearchAbout iterate.

(* Prove simultaneaously S <= ackermann i and strict_mono (ackermann i) *)








Lemma L45:  forall i, i < ackermann i 1 .
induction i. 
cbn. auto with arith. 
rewrite L44.
destruct (Ack_gt i).
specialize (H 1).
specialize (H0 _ _ H).
omega. 
Qed.




(*
Lemma L460 : forall i,  dominates_from 1 (ackermann (S (S i))) (f_ i).
induction i. 
  intro j.
  cbn. 
  admit.
  intro j.   
   rewrite F_S_rw; unfold Fsuc.
   rewrite Ack_S_rw.
  intro.    
 rewrite (iterate_S_eqn2  (ackermann (S (S i)))).
 rewrite iterate_S_eqn2. 
rewrite Ack_S_rw2.
Admitted. 

Lemma tortue  f g (H: forall i, 1 <= i ->  g i < f i) :
  forall k, 1 <= k -> iterate g k k < iterate f (S k) 1.
Proof. 
  induction 1.
  repeat rewrite iterate_S_eqn. 
  cbn. 
  admit. 
    repeat rewrite iterate_S_eqn. 
 *)

