(** Pierre Castéran, University of Bordeaux and LaBRI *)



Require Import Hydra_Games WfVariant Arith Omega.
Require Import TooSimpleVariant.


(* First try *)

Instance height_var : WfVariant round Peano.lt hheight.
Proof.
  split.  
  - apply lt_wf.
  - intros h h' H.
(*
  1 focused subgoal
(unfocused: 0), subgoal 1 (ID 9)
  
  h, h' : Hydra
  H : h -1-> h'
  ============================
   hheight h' < hheight h
 *)

Abort.


Lemma hheight_bad :  ~ WfVariant round Peano.lt hheight.
Proof.
  destruct 1 as [_ H].
  assert (H0 : hyd1 (hyd2 head head) -1->  hyd1 (hyd1 head)).
  { exists 0. right.  R2_here 0.  left.  }
  specialize (H _ _ H0);  compute in H;  omega.
Qed.

(** There is no measure into omega for proving termination
of all hydra battles *)

Section nat_too_simple.
  
  Let h := hyd1 (hyd1 head).
  Let iota (i: nat) := hyd_mult head (S i).


  (** Let us assume there is a variant from [Hydra] into nat for proving
     termination of all hydra battles *)
  
  Section hypothetic_variant. 
    Variable m : Hydra -> nat.
    Context (V : WfVariant round Peano.lt m). 



    Let h' := iota (m h).
    
    
    Lemma round_0 : h -1-> h'.
    Proof.
      exists (m h); right;  repeat constructor.     
    Qed.
    
    (**  In order to find a contradiction, we plan to prove the inequality
       m h' >= m h , i.e.  m (s (m h)) >= m h
      
       For that purpose, we prove the inequality i <= m (s i) for any i 
     *)
    
    Lemma round_S: forall i, iota (S i) -1-> iota i.
    Proof.
      intro i; exists 0; constructor 1; constructor;  induction i.
      - right;left.
      - cbn;  constructor 2; trivial.
    Qed.


    

    Lemma m_ge : forall i:nat , i <= m (iota i).
      induction i.
      - auto with arith.
      -  apply Lt.le_lt_trans with (m (iota i)).
         + assumption.
         + destruct V; apply decr, round_S.
    Qed.

  End hypothetic_variant.

  Lemma le_is_clos_refl : forall x y, le x y <->
                                      Relation_Operators.clos_refl _ lt x y.
  Proof.
    split; intros.
    -          destruct (le_lt_or_eq _ _ H).
               + now left.
               + subst; now right.
    - 
      destruct H;auto with arith.
  Qed. 


  Instance omega_too_small : TooSimple nat lt .
  Proof.
    exists iota h.
    - intros.
      
      rewrite <- le_is_clos_refl.
      apply m_ge .
      auto.
     -  left.
      apply round_0.
  Qed.
  
End nat_too_simple.


