Require Export List Arith Relations Omega.
Require Import Sorting.Sorted.
 Require Import Compare_dec.

(** numbers from i to i+n-1  *)

Fixpoint first_nats_aux i n :=
  match n with 0 => nil
            | S p =>  i :: first_nats_aux (S i) p
  end.


Definition first_nats n := first_nats_aux 0 n.

Example first_nats5 : first_nats 5 = 0::1::2::3::4::nil.
Proof. reflexivity. Qed. 

Definition interval i j := first_nats_aux i (S j - i).

(** Old APL's iota *)

Definition iota j := interval 1 j.

Example int_5_9 : interval 5 9 = 5::6::7::8::9::nil.
Proof. reflexivity. Qed. 

(** simpler than StdLib's last *)

Fixpoint simple_last s (d:nat) :=
  match s with
      nil => d
    | i::s' => simple_last  s' i
  end.


(** sorted list of elements greater or equal than n *)

Inductive sorted_ge (n: nat) : list nat -> Prop :=
| sorted_ge_nil : sorted_ge n nil
| sorted_ge_one : forall p, n<=p ->
                              sorted_ge n (p::nil)
| sorted_ge_cons: forall p q s,  n<=p -> p<q ->
                                   sorted_ge p (q::s) ->
                                   sorted_ge n (p::q::s).




Hint Constructors sorted_ge.
Require Import Sorting.Sorted.


Lemma sorted_inv_gt : forall n p s, sorted_ge n (p::s) ->
                                    (p < n)%nat -> False.
Proof.
 inversion_clear 1 ; intro;omega. 
Qed.


Section Forall2_right_induction.

Inductive Forall2R {A B: Type} (R: A -> B -> Prop) : list A -> list B -> Prop :=
  Forall2R_nil : Forall2R R nil nil
| Forall2R_last : forall l l' x y l1 l'1, Forall2R R l l' ->
                                          R x y ->
                                          l1 = l++(x::nil) ->
                                          l'1 = l' ++ (y::nil) ->
                                          Forall2R R l1 l'1.

Remark Forall2R_cons {A B: Type} (R: A -> B -> Prop):
  forall l l',  Forall2R R l l' -> forall x y, R x y -> Forall2R R (x::l) (y:: l').
  induction 1.
  - intros.
    eright with nil nil x y ; auto.
    left.          
  -  intros; subst;  right with (x0::l) (y0::l') x y; auto.
Qed. 

     
Remark Forall2_R  {A B: Type} (R: A -> B -> Prop) :
  forall l l', Forall2 R l l' -> Forall2R R l l'.
  induction 1.
  - constructor. 
  - inversion IHForall2.
     + eright with nil nil x y;auto.
       now left.     
     + subst; apply Forall2R_cons; auto.
Qed.


Remark Forall2_RR  {A B: Type} (R: A -> B -> Prop) :
  forall l l', Forall2R R l l' -> Forall2 R l l'.
Proof.
  induction 1.
  - constructor. 
  - subst; apply Forall2_app; auto.
Qed.

Lemma Forall2R_iff {A B: Type} (R: A -> B -> Prop) :
  forall l l', Forall2R R l l' <-> Forall2 R l l'.
Proof. 
  split.
  - intro;now  apply Forall2_RR .
  - intro;now  apply Forall2_R .
Qed. 


Lemma Forall2_indR {A B : Type} (R : A -> B -> Prop) (P : list A -> list B -> Prop):
P nil nil ->
(forall (l : list A) (l' : list B) (x : A) (y : B) 
   (l1 : list A) (l'1 : list B),
 Forall2 R l l' -> P l l' ->
 R x y -> l1 = l ++ x :: nil -> l'1 = l' ++ y :: nil -> P l1 l'1) ->
  forall (l : list A) (l0 : list B), Forall2 R l l0 -> P l l0.
Proof. 
  intros; rewrite  <- Forall2R_iff in  H1.
  eapply Forall2R_ind; eauto.
  intros; subst.
  eapply H0 with l1 l' x y; auto.  
  now  rewrite  <- Forall2R_iff .
Qed.    

End  Forall2_right_induction.

Definition ptwise_le: list nat -> list nat -> Prop := Forall2 le.




Lemma sorted_le : forall i j X, i <= j ->
                                sorted_ge j X ->
                                sorted_ge i X.
Proof.
  induction 2; eauto with arith. 
Qed.



  
Lemma sorted_tail : forall i j X, sorted_ge i (j::X) ->
                                  sorted_ge i X.
Proof.   
  inversion 1.
  -   constructor.
  -   destruct s; auto. 
      all:  inversion H4;eauto with arith.
Qed.

Lemma sorted_tail' : forall i j X, sorted_ge i (j::X) ->
                                   sorted_ge j X.
Proof.
  inversion_clear  1; auto.
Qed. 

Lemma sorted_head : forall n m s, sorted_ge n (m::s)
                                  -> n<=m.
Proof.
  induction s. 
  - inversion 1; auto. 
  - inversion_clear 1.
    apply IHs.
  inversion_clear H2.   
  +   constructor; auto with arith.
  +  constructor; eauto with arith.
     eapply sorted_le;eauto.
Qed.

Lemma sorted_le' : forall i j  X, 
                     sorted_ge i (j::X) -> forall k, k <= j ->
                                                       sorted_ge k (j::X).
Proof.
  induction X.
  -  auto with arith. 
  -  inversion_clear  1; auto with arith.  
Qed.


Lemma Sorted_mono {A:Type}(R S : relation A)
      (Hincl : forall x y, R x y -> S x y):
  forall l, Sorted R l -> Sorted S l.
Proof.   
  induction 1; auto. 
  constructor; auto. 
  destruct H0;auto. 
Qed.



Lemma sorted_ge_iff0 : forall l n, sorted_ge n l <->
                                    LocallySorted Peano.lt l /\
                                    List.Forall (fun i => n <= i) l.
Proof.
  induction l.
  - split; auto.
    split; auto; constructor.
  - destruct l.    
    + split.
      * split; auto.
        constructor.
        constructor.        
        eapply  sorted_head; eauto.
        constructor.    
      *   destruct 1;  constructor;  inversion H0;auto.
    +  repeat  split.
      *  inversion_clear H; auto.     
         rewrite IHl in H2.
         constructor;tauto.
      *   inversion_clear H.
          rewrite IHl in H2.
          constructor; auto.
          destruct H2.
          eapply Forall_impl.
          2:eapply H2.
          intros;   transitivity a.
          auto. 
          apply H3.
      *  destruct 1.
         constructor.
           inversion_clear H0; auto. 
           inversion_clear H; auto.
           inversion_clear H.
           rewrite IHl.
           split; auto.
           constructor;  auto with arith.
           apply Sorted_extends.
           intros x y z Hxy Hyz.
           eauto with arith.
           apply Sorted_mono with Peano.lt.
           eauto with arith. 
           rewrite <- Sorted_LocallySorted_iff in H1.
           inversion_clear H1.
           constructor;auto.
           induction H3.
           constructor.
           constructor. 
           eauto with arith. 
Qed. 

Lemma sorted_ge_iff : forall l n, sorted_ge n l <->
                                    Sorted lt l /\
                                    List.Forall (fun i => n <= i) l.
Proof.  
 intros.
 rewrite Sorted_LocallySorted_iff.
 apply sorted_ge_iff0.
Qed.


Lemma sorted_ge_prefix :
  forall  l1 n l2, sorted_ge n (l1 ++ l2) -> sorted_ge n l1.
  induction l1.
  - constructor.
  - 
    destruct l1.
    constructor.
    cbn in H.
    eapply sorted_head; eauto.
    constructor.
    cbn in H.
    eapply sorted_head; eauto.
    inversion_clear H.
    auto. 
    eapply IHl1.
    inversion_clear H.
    eauto. 
Qed. 


Lemma Sorted_In : forall i X, Sorted lt (i::X) -> forall j, In j X -> i < j.
Proof.
  intros.
  apply Sorted_StronglySorted in H.
  apply StronglySorted_inv in H.
  destruct H.
  rewrite Forall_forall in H1.
  auto.
  red. 
  intros; omega.
Qed.

Lemma sorted_not_in_tail : forall  i j X, Sorted lt (i::X)   -> j<=i ->
                                           ~ (In j X).
Proof.
 intros. 
 red; intro. 
  specialize (Sorted_In _ _ H j H1).
 intro; omega. 
Qed.

Remark simple_last_correct : forall s d, simple_last s d = last s d.
  induction s. 
  - now cbn.
  -  cbn.   
     destruct s.
     now cbn.
     auto.
Qed.

Lemma In_sorted_ge_inv : forall x y s,
                             In x (y::s) ->
                             sorted_ge y s ->
                             y < x /\ In x s \/ y = x.
Proof.
  intros x y s H H0;  destruct H.
  -    subst; auto. 
  -  destruct (lt_eq_lt_dec x y) as [[H1 | H2] | H3].
   +  rewrite   sorted_ge_iff in H0;  destruct H0.
     rewrite Forall_forall in H2;  specialize (H2 x H);  elimtype False.
     omega.
   +   subst; auto. 
   +  auto. 
Qed. 


Lemma incl_inv : forall x y l1 l2,  Sorted lt (x::l1) ->
                                    Sorted lt (y::l2)  ->
                                    incl (x::l1)(y::l2) ->
                                    y <= x /\ incl l1 l2.
Proof. 
  intros.
  red in H1.
  assert (y <= x).
 {
   specialize (H1 x (in_eq x  l1)).   
  destruct H1.
  subst; auto with arith.   

      specialize (Sorted_In _ _ H0 _ H1).
  auto with arith.
  }   
 split;auto.
  intros z Hz.
  assert (x < z). {

    specialize (Sorted_In _ _ H _ Hz).
    auto.
    }
 assert (y < z).
 eauto with arith. 

  specialize (H1 z).
   destruct H1.
   right;auto.
    subst.
   elimtype False.
 omega. 
 auto.
Qed.


Lemma incl_decomp : forall l1 l2,  Sorted lt l1 ->
                                    Sorted lt l2  ->
                                    incl l1 l2 ->
                                    exists l3 l4,  l2 = l3 ++ l4 /\
                                                   ptwise_le l3 l1.
induction l1.
 intros; exists nil, l2.
 split.
   reflexivity.
   constructor.
   destruct l2.
   intros.
   red in H1.
   destruct (H1 a) ;auto.
   now left.
   intros.   
  destruct (incl_inv a n l1 l2);auto.
  destruct (IHl1 l2).
  now inversion_clear H.
 now inversion_clear H0.
 auto. 
 destruct H4 as [l4 [H5 H6]].
  subst.
exists (n::x), l4.
 split;auto. 
 constructor.
 auto.
 
auto.
Qed.


  Lemma simple_last_app : forall l l1 x d, simple_last (l++(x::l1)) d =
                                     simple_last l1 x.
  induction l; cbn.
  - reflexivity.     
  - intros; now rewrite IHl.
Qed.


Lemma simple_last_app1 : forall l x d, simple_last (l++(x::nil)) d = x.
Proof.
  intros; now rewrite simple_last_app. 
Qed.


Lemma sorted_max_1 : forall s n, sorted_ge n s ->
                                 (n <= simple_last s n)%nat.
  induction s; cbn.
  - auto with arith. 
  -  intros; destruct s.
    + now inversion_clear H.
    + inversion_clear H; transitivity a; auto.  
Qed.


Lemma sorted_cut :forall l1 n x l2, sorted_ge n (l1++(x::l2)) ->
                                    simple_last l1 n <= x.
  induction l1.
  - cbn;  eapply sorted_head; eauto. 
  - cbn; intros; eapply IHl1 with l2.
    cbn in H;    inversion_clear H; auto.
Qed.

Lemma sorted_max_2 : forall s n, sorted_ge n s ->
                                 Forall (fun i =>
                                           (i <= simple_last s n)%nat)
                                        s.
Proof.
  induction s; cbn.
  - intros; simpl; constructor.
  -     constructor.
       +  apply sorted_max_1; inversion H;auto. 
       + apply IHs; eapply sorted_tail'; eauto.
Qed. 


Lemma sorted_ge_suffix :
  forall  l1 n l2, sorted_ge n (l1 ++ l2) -> sorted_ge (simple_last l1 n) l2.
  induction l1.
  -  intros; cbn. assumption. 
  -  destruct l1.
    + cbn; inversion_clear 1; auto.
    + inversion_clear 1; cbn in *; eauto.
Qed. 

