(* Pierre Casteran ,LaBRI et Projet Logical (Inria Futurs) *)


(* Well ordered sets *)
(*
Add LoadPath "../prelude".
Add LoadPath "../hilbert".
Add LoadPath "../denumerable".
 *)

Require Import Relations.
Require Import Classical.
Require Import Classical_sets.
Require Import RelationClasses.
Require Import PartialFun.

Arguments In [U].
Arguments Included [U].

Set Implicit Arguments.
Hint Unfold In.


Definition is_least_member{M:Type}(D:Ensemble M) (Lt : M -> M -> Prop)
             (X:Ensemble M)(a:M) :=
  In D a /\ In X a /\ forall x, In D x -> In X x -> x=a \/ Lt a x.

  
(** Well Ordering *)

Class WO(M:Type)(Lt : relation M)(D:Ensemble M): Type:= {

 Lt_inc_D1 : forall x y, Lt x y -> In D x;
 Lt_inc_D2 : forall x y, Lt x y -> In D y;
 Lt_trans : Transitive  Lt;
 Lt_irreflexive : forall a:M, ~ (Lt a a);
 well_order : forall (X:Ensemble M)(a:M),
                In D a -> In X a ->
                exists a0:M, is_least_member D  Lt X a0
}.

Hint Resolve Lt_inc_D1 Lt_inc_D2.

(* Some derived properties of well ordered sets *)

Section About_WO.

  Variables (M:Type)(Lt : relation M)(D:Ensemble M).
  Context (Wo : WO  Lt D).
 
  Definition Le (a b:M) := (In D a /\ a = b) \/ Lt a b.

  Lemma  Lt_connect : forall a b, In D a -> In D b -> Lt a b \/ a = b \/ Lt b a.
  Proof.
  intros a b H H0; generalize (well_order  (Couple _ a b) a H).
  destruct 1 as [ c H1].
  - left. 
  -  destruct H1 as [H2 [H3 H4]]; destruct H3.
      destruct (H4 b);auto; now right.
      destruct (H4 a);auto; now left.       
Qed.
 
 Lemma Le_refl : forall x, In D x -> Le x x.
 Proof.
  red;unfold Le;auto.
 Qed.

 Lemma Le_antisym : forall a b,  Le a b -> Le b a -> a = b.
 Proof.
  intros a b  H H'; case H; case H' ; try tauto.
  - symmetry; tauto.
  -   intros;case (Lt_irreflexive a); eapply (Lt_trans );eauto. 
 Qed.

 Global Instance  Le_trans : Transitive  Le. 
 Proof.
  unfold Transitive, Le;intros.
  case H;case H0.
  -   intuition.
      left;intuition.
      transitivity y;auto.
      left;split; auto.
      transitivity y;auto.
      left;split;  auto.
      transitivity y;auto.
      right;apply Lt_trans with y;eauto.
   -  intros H1 (H2,H3); right;subst x;auto.
   -  intros (H1,H2) H3; right ; subst y; auto.
   -  right;apply Lt_trans with y;eauto.
Qed.


 Lemma Le_Lt_trans : forall x y z, Le x y -> Lt y z -> Lt x z.
 Proof.
  intros x y z Hxy Hyz; case Hxy;intros.
  - destruct H;   now subst y.
  - eapply Lt_trans; eauto.
 Qed.

 Lemma Lt_Le_trans : forall x y z, Lt x y -> Le y z -> Lt x z.
 Proof.
  intros x y z Hxy Hyz.
  destruct Hyz as [[H H0] | H].
  - now subst.
  -  eapply Lt_trans; eauto.
 Qed.

 Lemma Lt_not_Gt : forall x y,  Lt x y -> ~ Lt y x.
 Proof.
  intros x y  H H'; case (Lt_irreflexive x).
  eapply Lt_trans; eauto.
 Qed.

 Lemma least_member_lower_bound : forall X a,
   is_least_member D Lt X a -> forall b, In X b -> In D b -> Le a b.
 Proof.
  intros X a H; case H.
  unfold In; intuition.
  unfold Le; case (H3 _ H4 H1);auto.
 Qed.

 Lemma least_member_glb :
   forall X a,
   is_least_member D Lt X a -> 
     forall b, (forall c, In X c -> In D c -> Le b c) ->
               Le b a.
 Proof.
   intros X a H b H0; case H;intros H1 (H2,_);  apply H0; auto.
 Qed.


 Theorem least_member_unicity : forall  X a b, 
    is_least_member D Lt X a -> is_least_member D Lt X b -> a = b.
 Proof.
  intros X a b H H0;  case H;case H0;intros.
  - apply Le_antisym;auto.
    case H2;intros H5 H6.
    case H4;intros H7 H8.
    red;  case (H8 b H1 H5);auto.
    red ;  case H2;intros H5 H6.
    case H4;intros H7 H8.
    case (H6 a H3 H7);auto.
 Qed.


 Theorem least_member_ex_unique :
   forall  (d:M) X (H:Included X D) x (inhD: D d)
           (inhX: In X x), 
    exists! a,  is_least_member D Lt X a.
 Proof.
   intros;destruct (well_order X x); auto.
   exists x0; split; auto.
   intros; eapply least_member_unicity;eauto.
Qed.

 
 Theorem least_member_of_eq : forall (X Y : Ensemble M) a b ,
                                Included X Y -> Included Y X ->
                                is_least_member D Lt X a -> is_least_member D Lt Y b ->
                                a = b.
 Proof.
   intros X Y a b H H0  [H3 [H4 H5]] [H6 [H7 H8]].  apply Le_antisym;auto.
   - destruct (H5 b);auto.
     subst;     left;auto.
     right;auto.     
   -  destruct (H8 a);auto.
      subst;     left;auto.
      right;auto.
 Qed.


Section restriction.

 Variable a : M.
 Hypothesis Ha : D a.
 Variable a0 : M.

 Definition D_a x := In D x /\ Lt x a.
 Definition Lt_a := fun x y => Lt x a /\ Lt y a /\ Lt x y.


 Remark Lta_inc_D1 :  forall x y, Lt_a x y -> D_a x.
 Proof.
   unfold D_a; destruct 1; split;auto.
  eapply Lt_inc_D1; eauto.
 Qed.

 Remark Lta_inc_D2 :  forall x y, Lt_a x y -> D_a y.
 Proof.
  unfold D_a;destruct 1; split;auto.
  - destruct H0 ;  eapply Lt_inc_D2; eauto.
  -  tauto.
 Qed.

 Remark Lta_irr : forall x,  ~ Lt_a x x.
 Proof.
  unfold Lt_a; intros x Hx; destruct (Lt_irreflexive x); try tauto.
Qed.


Remark Lta_trans :  transitive _ Lt_a.
Proof.
 unfold Lt_a; repeat split; try tauto.
 apply (Lt_trans  ) with y; tauto. 
Qed.


Remark Lta_connect : forall x y, D_a x -> D_a y -> 
                       Lt_a x y \/ x = y \/ Lt_a y x.
Proof.
 intros x y [H1 H2] [H3 H4].  
 case (Lt_connect  H1 H3 ).
 left;unfold Lt_a;auto.
 destruct 1.
 +  auto.
 +  right;right;unfold Lt_a;auto.
Qed.

Remark well_order_a : forall (M0:M->Prop)(m:M), 
                        D_a m -> M0 m ->
                        exists a0:M, is_least_member  D_a Lt_a M0 a0.
Proof.
  intros. 
  case (well_order  (fun y => M0 y /\ Lt y a) m).
  -  case H;auto.
  -  case H;auto.
  -  intros.
     assert (H2:In  D x) by (now destruct H1).
     case (Lt_connect H2 Ha ).
     + (* x < a *)
       exists x.      
       repeat split;auto.
       { destruct H1.
         destruct H4.
         destruct H4;  auto.
       }
       intros;  destruct H1.
       destruct H6.
       destruct (H7 x0).
       now destruct H4.
       split;  auto.
       now destruct H4.
       now left.
       right;auto.
       split;auto.
       split;auto.
       destruct H4;auto.
     +  destruct 1.
        * (* x = a*)
          subst x.
          destruct H1.
          destruct H3.
          destruct H3.
          destruct (Lt_irreflexive a H5).
        * (* a < x *)
          destruct (Lt_irreflexive a ).
          apply Lt_Le_trans with x.
          auto.
          unfold is_least_member in H1;decompose [and] H1.
          right.
          destruct H6;auto.
Qed.


Definition restrict : WO  Lt_a D_a.
Proof.
 split.
 apply Lta_inc_D1.
 apply Lta_inc_D2.
 apply Lta_trans.
 apply Lta_irr.
 intros; destruct  (@well_order_a X a1);auto.
 now exists x.
Defined.

End restriction.

End About_WO.

Section isotonicity_def.
  Variables (M1 M2 : Type) (Lt1 : relation  M1) (Lt2 : relation M2)
            (D1 : Ensemble M1)(D2 : Ensemble M2).
 Context (W1 : WO Lt1 D1) (W2 : WO Lt2 D2).

 Section f_given.

 Variable f : M1 -> M2 -> Prop.

 Definition rel_mono := forall a1 b1 a2 b2, In D1 a1 -> In D1 b1 ->
      f a1 a2 -> f b1 b2 -> Lt1  a1 b1 -> Lt2  a2 b2.

 Definition rel_isotonic := rel_domain D1    f /\
                            rel_functional D1    f /\ 
                            rel_onto D1 D2 f/\
                            rel_codomain D1 D2 f /\
                            rel_mono.


 Lemma iso_inj : rel_isotonic -> rel_inj D1  f.
 Proof.
  unfold rel_inj, rel_isotonic;intro H;decompose [and] H.
  intros a a' b H4 H6 H7 H8.
  case (Lt_connect W1 a a');auto.
  intro H9; destruct (Lt_irreflexive  b).
  apply H5 with (a1:=a)(b1:=a');auto.
  destruct 1;auto.
  destruct (Lt_irreflexive  b).
  apply H5 with (a1:=a')(b1:=a);auto.
 Qed.

End f_given.

Definition isotonic := exists f, rel_isotonic f.

End isotonicity_def.

Theorem isotonicity_refl : forall M (Lt:relation M)  D,  (isotonic Lt Lt D D).
Proof.
 unfold isotonic; intros.
 exists (fun (z z':M ) => z=z').
 unfold rel_isotonic.
 repeat split.
 unfold rel_domain;intros.
 exists a;auto.
 unfold rel_functional;intros.
 subst b;auto.
 unfold rel_onto;intros.
 exists b;auto.
 unfold rel_mono;intros.
 red.
intros; now subst.
 red.
 intros; now subst. 
Qed.

(*  UNUSED 
Theorem isotonicity_sym : symmetric  _ isotonic.
Proof.
 unfold isotonic, symmetric; intros.
 case H;intros f Hf.
 exists (fun z z' => D x z' /\ f z' z).
 unfold rel_isotonic in Hf;decompose [and] Hf.
 unfold rel_isotonic.
 repeat split.
 unfold rel_domain;intros.
 case (H1 _ H4).
 intros x0 H';exists x0;auto.
 unfold rel_functional;intros.
 generalize (iso_inj (f:=f)); unfold rel_inj; intro Inj.
 eapply Inj. 
 split; auto.
 tauto.
 tauto.
 instantiate (1:=a).
 tauto.
 tauto.
 unfold rel_onto.
 intros.
 case (H0 b).
 auto.
 intro x0;exists x0.
 generalize (H3 b x0).
 tauto.
 unfold rel_codomain;intros.
 tauto.
 unfold rel_mono;intros.
 case (Lt_connect x a2 b2). 
 tauto.
 tauto.
 auto.
 destruct 1.
 subst b2.
 replace a1 with b1 in H9.
 case (Lt_irreflexive y b1);auto.
 eapply (H2 a2 b1 a1).
 tauto.
 tauto.
 tauto.
 assert (Lt y b1 a1).
 eapply (H5 b2 a2).
 tauto.
 tauto.
 tauto.
 tauto.
 auto.
 case (Lt_irreflexive y a1).
 eapply (Lt_trans y);eauto.
Qed.
 *)

(*
 
Theorem isotonicity_trans : transitive  _ isotonic.
Proof.
 unfold isotonic, transitive; intros.
 case H;intros f Hf.
 case H0;intros f' Hf'.
 exists (fun z z'' => D x z /\ (exists z', f z z' /\ f' z' z'')).
 unfold rel_isotonic in Hf;decompose [and] Hf.
 unfold rel_isotonic in Hf';decompose [and] Hf'.
 unfold rel_isotonic.
 repeat split.
 unfold rel_domain;intros.
 case (H1 _ H10).
 intros x0 H'.
 case (H5 x0).
 apply (H4 a).
 auto.
 auto.
 intros.
 exists x1.
 split.
 auto.
 exists x0;auto.
 unfold rel_functional;intros.
 case H12;intros.
 case H13;intros.
 case H15;intros;case H17;intros.
 assert (x0=x1).
 apply (H3 a x0 x1).
 auto.
 tauto.
 tauto.
 subst x1.
 apply (H8 x0 b b').
 apply (H4 a).
 auto.
 tauto.
 tauto.
 tauto.
 unfold rel_onto.
 intros.
 case (H7 b).
 auto.
 intros x0 (H12,H13). 
 case (H2 x0).
 auto.
 intros x1 (H14,H15).
 exists x1.
 repeat split;auto.
 exists x0;auto.
 unfold rel_codomain;intros.
 case H12;intros H13 (z',(H14,H15)).
 apply (H9 z');auto.
 apply (H4 a);auto.
 unfold rel_mono;intros.
 case H13;intros H16 (z',(H17,H18)).
 case H14;intros H19 (z'',(H20,H21)).
 apply (H11 z' z'').
 apply (H4 a1);auto.
 apply (H4 b1);auto.
 auto.
 auto.
 apply (H6 a1 b1);auto.
Qed.

 *)
(*
Inductive WO_lt (W W1:WO) : Prop :=
WO_lt_i:  forall a (H:D W1 a), isotonic W (restrict W1 a H) ->
  WO_lt W W1.

 *)

