Require Import Logic.Description.


Section iota.
  Variables (A:Type)(P: A -> Prop).
  Hypothesis ex_unic : exists ! x : A, P x.

  Definition iota : A :=
    let s := constructive_definite_description P ex_unic in let (x, _) := s in x.

  Lemma iota_spec1 : P (iota).
    unfold iota.
    destruct (constructive_definite_description P ex_unic).
    assumption.
  Qed.

  Lemma iota_spec : unique P (iota).
    unfold iota.
    destruct (constructive_definite_description P ex_unic).
    destruct ex_unic.
    destruct H;split.
    specialize (H0 _ p).
    now subst x0.
    intros.
    transitivity x0.
    symmetry; auto.
    auto.
  Qed.

  Lemma iota_eq : forall a : A, P a -> a = iota.
     intros.
    destruct ex_unic.      
     red in H0.
     destruct H0.
     transitivity x.
     symmetry;auto.
     apply H1;apply iota_spec1.
    Qed.


End iota.


 

Arguments iota {A P} _.

Lemma iota_e : forall (A:Type) (a: A) (P : A -> Prop)
                     (H:ex (unique P)),
                     unique P (iota  H ).
Proof.
  intros A a P  H.
  unfold iota.
    destruct (constructive_definite_description P H).
    split; auto.
     destruct H;intros.
     destruct H;    transitivity x0;auto.
   symmetry;auto.
Qed.



Section Examples.

Let P (n:nat) := forall p, n <= p.

Lemma P_unique : exists! i, P i.
Proof.
  exists 0.
  split.
  red;auto with arith.
  unfold P;intros. specialize (H 0). now inversion_clear H.
Qed.

Let zero := iota P_unique.

Goal zero = 0.
  symmetry.
  apply iota_eq.
  red;intros.
  auto with arith.
Qed.

End Examples.