(* * A sorting example :  *)
(*    (C) Yves Bertot, Pierre Casteran  *)
(* *) 


Require Import Bool List  ZArith  Relations  RelationClasses Morphisms.
Require Import  Coq.Sorting.Permutation Coq.Sorting.Sorted  DecPreOrder.
Require Import DecPreOrder_Instances Sort_spec Extraction.
Set Implicit Arguments.

(** A small polymorphic sorting function *)

(** Definitions *)

Section A_fixed.
 Variable A : Type.

 Fixpoint insert (leb: A -> A -> bool)  (a:A)(l:list A) : list A :=
 match l with
   | nil => a::nil
   | b::l' => if leb a b then a::l else b::insert leb a l'
 end.

 Fixpoint insertion_sort (leb: A -> A -> bool) (l:list A) : list A :=
 match l with
   | nil => nil
   | a::l' => insert leb a (insertion_sort leb l')
 end.

Section Correctness. 
Variable le : relation A.
Context (le_total : TotalDecPreOrder le).

Notation "a <= b" := (le a b).
Notation "a <=? b" := (leb le a b).


Lemma insert_LocallySorted :
  forall (l:list A) (x:A), LocallySorted le l -> LocallySorted le (insert (leb le) x l).
Proof.
  intros l x H; induction  H ; simpl;  auto .
  -  destruct  (dec_dec x a).
     (* x <= a *)
     le_2_leb l H ; rewrite H; auto.
     (* ~ x <= a *)
     replace (x<=? a) with false.
     assert (a <= x) by ( now apply not_le_ge).
     auto.
     (** prove ~ x <= a -> x <=? a = false *)
     {  case_eq (x <=? a).
        intro e.
        leb_2_le e H;      contradiction. 
        auto.
     }
  - case_eq (x <=? a).    
    + intro H3;   leb_2_le H3 H4.
      auto.
    + 
      intro H3.
      destruct (dec_dec x b).
      le_2_leb l0 H1.       rewrite H1.
      assert (a <= x).
      { 
        rewrite   (@leb_false A _ le_total ) in H3.        
        now apply le_lt_weak.
      }
      auto.
      assert (x <=? b = false).
      rewrite   (@leb_false A _ le_total ).
      destruct (left_right_dec x b).
      destruct n.
      auto.
      split;auto.
      simpl in IHLocallySorted.
      rewrite H1 in *;auto.

Qed.


Lemma insertion_sort_sorts : forall l, LocallySorted le (insertion_sort (leb le)  l).
Proof.
  induction l; simpl insertion_sort;auto .
  apply insert_LocallySorted;auto.
Qed.


Lemma  permutation_transpose :
 forall (a b:A) (l l':list A),
   Permutation l l' -> 
   Permutation (a :: b :: l) (b :: a :: l').
Proof.
 intros a b l l' H .
 transitivity (a::b::l').
 now repeat constructor.
 apply perm_swap.
Qed.



Hint Resolve    permutation_transpose : sort.


Lemma insert_permutation : forall (x:A)(l:list A),
                             Permutation (x :: l) (insert (leb le) x l).
Proof with auto .
  induction l as [|a l0 H]; simpl...  
  destruct (left_right_dec x a);simpl ...
  transitivity (a :: x :: l0) ...
  now  rewrite permutation_transpose.
  rewrite H.
  le_2_leb l H0;  rewrite H0.
  auto.
  transitivity (a::x::l0).
  symmetry.
  constructor. auto.
  apply perm_swap.

  destruct (le_lt_equiv_dec le le_total a x );simpl ...
  replace (x <=? a) with false in *.
  transitivity (a::x::l0);auto.
  apply permutation_transpose.
  reflexivity. 
  symmetry;rewrite   (@leb_false).
  auto.
  destruct p.
  le_2_leb H1 H2;rewrite H2.
  reflexivity.
Qed.


Hint Resolve insert_permutation .

Lemma sort_permutation : forall l, Permutation l (insertion_sort  (leb le) l).
Proof with auto .
  induction l;simpl insertion_sort ...
  transitivity (a::insertion_sort  (leb le) l) ...
Qed.

Lemma sort_ok : sort_correct _  le  (insertion_sort (leb le) ).
Proof.
  split.
  - apply insertion_sort_sorts.
  -  apply sort_permutation.
Qed.

End Correctness.
End A_fixed.

Theorem insertion_OK : sort_spec insertion_sort.
Proof.
  split.
  apply sort_permutation.
  apply  insertion_sort_sorts.
Qed.



Require Import ZArith.
Open Scope Z_scope.

Locate sort.


Extraction "insertion-sort" insert insertion_sort.


Extraction Language Scheme.  
Extraction "insertion-sort" insert insertion_sort.



