(* (C) Pierre Castran ,Labri, Universit Bordeaux 1 *)

Require Import Arith  Epsilon  Ensembles.
Require Import Schutte  Ordering_Functions  PartialFun  Denumerable  MoreIota.
Set Implicit Arguments.


Definition GreaterEq alpha : Ensemble OT := fun beta => alpha <= beta.

Definition plus alpha := the_ordering_function (GreaterEq alpha).

Notation "alpha + beta " := (plus alpha beta) : ord_scope.

Lemma GreaterEq_inc_ord : forall alpha, ordinal alpha ->
                       Included (GreaterEq alpha) ordinal.
Proof with auto with schutte.
 intros alpha H x H0;  case H0.
 -  destruct 1; subst x;auto.
 - intro; red;apply gt_ordinal with alpha;auto.
Qed.

Hint Resolve GreaterEq_inc_ord.

Lemma Unbounded_GreaterEq : forall alpha0, ordinal alpha0 ->
        Unbounded (GreaterEq alpha0).
Proof with auto with schutte.
  intros alpha0 H x H0; tricho alpha0 x H3.
 -  exists (succ x);split ...
    right;  apply lt_trans with x; trivial ...
 -  subst alpha0; exists (succ x) ; split ...
    red;red ...
 - exists (succ alpha0); split.
  +  red;red ...
  +  apply lt_trans with alpha0 ...
Qed.

Lemma GreaterEq_o_segment : forall alpha0, ordinal alpha0 ->
 the_ordering_segment (GreaterEq alpha0) = ordinal.
Proof.
 intros;apply O_segment_unbounded.
 - pattern  (the_ordering_segment (GreaterEq alpha0)); apply iota_ind.
   +  generalize  (B_of_u (GreaterEq_inc_ord H)); destruct 1, H0.
      exists x; split;auto.
   +  destruct 1;eapply ordering_function_seg;eauto.
 -     generalize 
        (ordering_unbounded_unbounded 
           (A:=the_ordering_segment (GreaterEq alpha0)) 
           (B:=GreaterEq alpha0) 
           (f:=plus alpha0));  intros H0.
       generalize (H0 (GreaterEq_inc_ord H)).
       intro H1; clear H0.
       unfold plus; generalize (H1 (the_ordering_function_ok (GreaterEq_inc_ord H))).
       destruct 1; apply H0; apply Unbounded_GreaterEq;auto.
Qed.
 


Lemma plus_ordering : forall alpha, ordinal alpha ->
                                    ordering_function (plus alpha)
                                                      ordinal
                                                      (GreaterEq alpha).
Proof.
  intros alpha H; unfold plus;
  generalize (@the_ordering_function_ok (GreaterEq alpha));
    rewrite  GreaterEq_o_segment;auto.
Qed.

Lemma plus_elim : forall alpha,
                    ordinal alpha ->
                     forall P : (OT->OT)->Prop,
                     (forall f: OT->OT, 
                       ordering_function f ordinal (GreaterEq alpha)-> P f) ->
                      P (plus alpha).
Proof.
 intros alpha H P H0; now apply H0, plus_ordering.
Qed.

Lemma normal_plus_alpha :  forall alpha, ordinal alpha -> 
                           normal (plus alpha) (GreaterEq alpha).
Proof.
 intros alpha H; unfold plus;pattern (the_ordering_function (GreaterEq alpha)).
 unfold the_ordering_function, some;apply epsilon_ind. 
 -  rewrite GreaterEq_o_segment; trivial.
   +  exists (plus alpha);apply (plus_ordering H); auto.
-  intros a H0; rewrite GreaterEq_o_segment in H0; trivial.
   split; trivial.
   apply Th_13_5_2; trivial.
   split.
   +  intros beta Hbeta.
      red in Hbeta; red in Hbeta; red;eauto with schutte.
   +   intros M H1 H2 H3;  destruct H2; apply le_trans with x;auto.
       *  specialize (H1 _ H2);  apply H1.
       *  apply sup_upper_bound;auto. 
          intros x0 H4;   specialize (H1 _ H4); eauto with schutte.
Qed.

 
Fixpoint mult_Sn (alpha:OT)(n:nat){struct n} :OT :=
 match n with 0 => alpha
            | S p => mult_Sn alpha p + alpha
 end.


Lemma alpha_plus_zero : forall alpha, ordinal alpha -> alpha + zero = alpha.
Proof.
 intros alpha H; pattern  (plus alpha); apply plus_elim;eauto.
 intros f H0 ; case (order_function_least_least H0).
  intros H1 [H2 H3]; case (H3 alpha);auto.
 -  red;red; auto with schutte.
 -  intro H4; case H2.
  +  symmetry;tauto.
  +  intro; case (@lt_irr alpha);auto.
     apply lt_trans with (f zero);auto.
Qed.

Remark GreaterEq_zero : (GreaterEq zero : Ensemble OT) = ordinal.
Proof with eauto with schutte.
 apply Extensionality_Ensembles; split.
 +  apply GreaterEq_inc_ord ...
 + unfold GreaterEq ...
Qed.

Lemma zero_plus_alpha : forall alpha, ordinal alpha -> zero + alpha = alpha.
Proof with auto with schutte.
 intros alpha H; pattern (plus zero);  apply plus_elim ...
 intros f H0; assert (H1:ordering_function (fun o => o) ordinal ordinal).
 {  (repeat split) ...
     -  eauto with schutte.
     -  intros b Hb; exists b;auto.
 }
 rewrite GreaterEq_zero in H0.
 generalize (ordering_function_unicity H0 H1); destruct 1; auto.
Qed.


Lemma le_plus_l : forall alpha beta, ordinal alpha -> ordinal beta ->
                                     alpha <= alpha + beta.
Proof.
  intros; pattern alpha at 1;  rewrite  <- alpha_plus_zero;auto.
  pattern (plus alpha);apply plus_elim;auto.
  intros;eapply ordering_function_mono_weak;eauto with schutte.
Qed.



Lemma le_plus_r : forall alpha beta, ordinal alpha -> ordinal beta ->
                                     beta <= alpha + beta.
Proof.
 intros alpha beta H H0;  pattern (plus alpha). 
 apply plus_elim;auto.
 intros; eapply ordering_le; eauto.  
Qed.


Lemma plus_mono_r : forall alpha beta gamma, 
                        ordinal alpha -> 
                        beta < gamma -> 
                        alpha + beta < alpha + gamma.
Proof.
 intros alpha beta gamma H H0; pattern (plus alpha); apply plus_elim;auto.
 intros; eapply ordering_function_mono;eauto with schutte.
Qed.

Lemma ordinal_plus : forall alpha beta, 
                       ordinal alpha -> 
                       ordinal beta ->
                       ordinal (alpha + beta).
Proof.
 intros; pattern (plus alpha); apply plus_elim; auto with schutte.
 intros; eapply ordering_function_ordinal_B;eauto.
Qed.

Hint Resolve ordinal_plus : schutte.
  
Lemma plus_mono_r_weak :forall alpha beta gamma, 
                        ordinal alpha -> 
                        ordinal beta ->
                        ordinal gamma ->
                        beta <= gamma ->
                        alpha + beta <= alpha + gamma.
Proof.
  intros alpha beta gamma H H0 H1 H2;  case (le_disj H2).
 - intros; subst gamma; auto with schutte.    
 - right;  apply plus_mono_r;auto.
Qed.


Lemma plus_reg_r : forall alpha beta gamma, ordinal alpha ->
                                            ordinal beta ->
                                            ordinal gamma ->
                                            alpha + beta = alpha + gamma  ->
                                            beta = gamma.
Proof with trivial.
 intros alpha beta gamma H H0 H1 H2;   tricho beta gamma H3 ...
 -  case (@lt_irr (alpha+beta)).
    pattern (alpha+beta) at 2; rewrite H2.
    apply plus_mono_r;auto.
 -  case (@lt_irr (alpha+beta)).
    pattern (alpha+beta) at 1; rewrite H2.
    apply plus_mono_r;auto.
Qed.


Lemma plus_of_succ : forall alpha , ordinal alpha ->
                                    forall beta, ordinal beta ->
                                          alpha + (succ beta) = succ (alpha + beta).
Proof with trivial.
 intros alpha H beta H0; pattern (plus alpha); apply plus_elim ...
 intros plus_alpha Hp;
   assert (H1 :plus_alpha beta < plus_alpha (succ beta)).
 {  eapply ordering_function_mono;eauto.
    red;auto with schutte.
     apply lt_succ; auto.
 }
 generalize (@lt_succ_le (plus_alpha  beta) (plus_alpha  (succ beta))).
 intros H2;  assert (H3 : ordinal (plus_alpha beta)).
 { case Hp; intros H3 H4; decompose [and] H4.
   apply H5.
   apply H7;auto.
 }
 assert (H4: ordinal (plus_alpha (succ beta))) by
     (eauto with schutte).
 generalize (H2 H1);  intro H5; case (le_disj H5) ; auto. 
 intro H6;case Hp;intros H7 H8; decompose [and] H8.
 case (H10 (succ (plus_alpha beta))).
 -  red; apply le_trans with (plus_alpha  beta).
   +  generalize (H11 beta H0).
      unfold In,  GreaterEq;  auto.
   +  right; apply lt_succ;  auto.
 -  intros x [Hx Ex]; clear H8; absurd ( beta < x /\ x < succ beta).
    + intro H8; decompose [and] H8;clear H8.
      assert (H8: succ beta <= x) by (apply lt_succ_le; eauto with schutte).
      case (@lt_irr (succ beta)).
      eapply le_lt_trans;eauto.
    +  split.
   *  eapply ordering_function_monoR;eauto.
      rewrite Ex; auto with schutte.
  *  rewrite <- Ex in H6; eapply ordering_function_monoR;eauto.
     auto with schutte.
Qed.

Lemma succ_is_plus_1 : forall alpha, ordinal alpha -> succ alpha = alpha + F 1.
Proof. 
  intros; rewrite <- (alpha_plus_zero H) at 1.
  rewrite <- plus_of_succ.
   f_equal.
 auto. 
 auto with schutte.  
Qed. 


Lemma alpha_plus_sup : forall alpha A , ordinal alpha -> Inhabited _ A ->
                                 denumerable A ->
                                 Included A ordinal ->
                                 alpha + |_| A = |_| (image A (plus alpha)).
Proof.
 intros alpha A H H0 H1 H2;  generalize (@normal_plus_alpha alpha H).
  intro H3;  case H3;  destruct 2.
  destruct H6 as [H7 H8];  now rewrite <- (H8 A H2 H0 H1).
Qed.


Lemma plus_limit : forall alpha beta, is_limit beta ->
                                      ordinal alpha ->
                                      alpha + beta =
                                      |_| (image (members beta) (plus alpha)).
Proof.
 intros alpha beta H H0;  generalize (is_limit_ordinal H); intro H1.
 generalize (is_limit_sup_members H);intro e.
 generalize (@normal_plus_alpha alpha H0); intro H2.
 pattern beta at 1;rewrite e; case H2; destruct 2, H5.
 rewrite <- H6;auto.
 -  red;destruct 1;auto.
 - case H; destruct 2; exists zero.
   split; auto with schutte.
   generalize (le_zero  H1).
   intro H10; case (le_disj H10).
   +  intro; subst beta; case H8;auto.
   + now  red.
 - apply denumerable_members; auto.
Qed.


Lemma plus_FF : forall i j, F (i + j) = F i + F j.
Proof. 
 induction i.
 -  simpl.
     intros;rewrite zero_plus_alpha;auto with schutte.
 -  simpl; induction j.
    +  simpl.
       rewrite alpha_plus_zero;auto with schutte.
       now rewrite <- (plus_n_O i).
    +  rewrite IHi.
       replace (F (S j)) with (succ (F j)).
       * rewrite plus_of_succ, plus_of_succ, <- IHi, IHj; auto with schutte.
       * now simpl.
Qed.


Lemma one_plus_omega : F 1 + omega = omega.
Proof.
 rewrite plus_limit; auto with schutte.
 - unfold omega at 2; apply le_antisym; auto with schutte.
  +    unfold omega_limit;  apply sup_mono;auto with schutte.
   *   apply Ordering_Functions.R1 with ordinal (GreaterEq (F 1)).
       apply plus_ordering, ordinal_finite.
       apply    denumerable_members;auto with schutte.
       red; destruct 1;  eauto with schutte.
   * red; destruct 1; decompose [and] H; subst x.
     apply ordinal_plus; auto with schutte.
     destruct H0; auto with schutte.
   * red. destruct 1.  subst x;apply ordinal_finite.
   *      intros x [a [H1 H2]]; generalize H2;case H1; intros.
          red in H0; case (@lt_omega_finite _  H0).
          intros n e; exists (F (S n)).
          subst x0; split;auto with schutte.
          exists (S n); trivial.
          rewrite <- plus_FF in H3. 
          left; split;auto.
          subst x;  red;apply ordinal_plus; auto with schutte.
          case H1;auto.
 +  unfold omega_limit; apply sup_mono.
  *    apply seq_range_denumerable; auto with schutte.
  *  simpl; auto with schutte.
     apply Ordering_Functions.R1 with ordinal (GreaterEq (succ zero)).
     apply plus_ordering; auto with schutte.
     apply denumerable_members; auto with schutte.
     red; destruct 1; red in H0; eauto with schutte.
  *  red;destruct 1.
     subst x; red;auto with schutte.
  * red; destruct 1; decompose [and] H.
    red;subst x;apply ordinal_plus;auto with schutte.
    case H0;auto.
  *  intros x Hx;case Hx; intros x0 H0;subst x.
     exists (F (S x0));split;auto with schutte.
     exists (F x0);split;auto with schutte.
     red.  red. split.  red. apply ordinal_finite.
     apply finite_lt_omega.
     rewrite <- plus_FF; trivial.
     
 - apply is_limit_omega.
Qed.


Lemma minus_exists : forall alpha beta, alpha <= beta ->
                                        exists gamma, ordinal gamma /\
                                                      alpha + gamma = beta.
Proof.
 intros alpha beta H.
 assert (H0 : ordinal alpha) by ( eauto with schutte).
 assert (H1 : ordinal beta) by ( eauto with schutte).
 pattern (plus alpha); apply plus_elim;auto.
 intros f H2; case H2;intros H3 H4; decompose [and] H4; clear H4.
 case (H6 _ H); intros;exists x;auto.
Qed.


Section proof_of_associativity.
 Variables alpha beta : OT.
 Hypothesis O_alpha : ordinal alpha.
 Hypothesis O_beta : ordinal beta.

Lemma plus_assoc1 : forall  gamma,
                       ordinal gamma ->
                       alpha + beta <= alpha + (beta + gamma) .
Proof.
 intros;apply plus_mono_r_weak;auto.
 -  apply ordinal_plus;auto.
 -  apply le_plus_l; auto.
Qed.

Lemma plus_assoc2 :  forall  gamma,
                       ordinal gamma ->
                       alpha + beta <= gamma ->
                       exists khi, ordinal khi /\ gamma = alpha + (beta + khi).
Proof.
 intros gamma H H0; assert (H1 : alpha <= gamma).
 {  apply le_trans with (alpha + beta);auto.
    apply le_plus_l; auto.
 }  case (minus_exists H1); intros z [Hz Hz'].
    assert (H2 :beta <= z).
   {  tricho beta z HH.
      -  right;auto.
      -  left;auto.
      -  case (@lt_irr gamma).
         pattern gamma at 1; rewrite <- Hz'.
         apply lt_le_trans with (alpha + beta);auto.
         + apply plus_mono_r;auto.
    }
   case (minus_exists H2);intros u (Hu,Hu').
   exists u;split;auto.
   rewrite Hu';auto.
Qed.

Let f_alpha_beta := plus (alpha + beta).

Let g_alpha_beta gamma := alpha + (beta + gamma).

Remark of_g : ordering_function g_alpha_beta ordinal (GreaterEq (alpha+beta)).
Proof.
 repeat split.
 -  red;auto.
 -  eauto with schutte.
 -  red;  destruct 1; eauto with schutte.
    decompose [and] H; subst x; eauto with schutte.
 - intros;red; red; unfold g_alpha_beta; now apply plus_assoc1.
 -   intros b H; assert (H0 :ordinal b) by ( eauto with schutte).
     case (plus_assoc2 H0 H); intros x [Hx H'x];  exists x.
    unfold g_alpha_beta;split;auto.
-  unfold g_alpha_beta;intros.
  + apply plus_mono_r;auto with schutte.
    apply plus_mono_r;auto with schutte.
Qed.

Lemma of_u : fun_eq_gen f_alpha_beta g_alpha_beta ordinal ordinal.
Proof.
 eapply ordering_function_unicity.
 - unfold f_alpha_beta; apply plus_ordering.
   apply ordinal_plus;auto.
 - apply of_g.
Qed.

Lemma plus_assoc3 : forall gamma, ordinal gamma -> f_alpha_beta  gamma =
                                                   g_alpha_beta gamma.
Proof.
  intros gamma H; case of_u; intros H0 H1; auto.
Qed.

Lemma plus_assoc' : forall gamma, ordinal gamma ->
                        alpha + (beta + gamma) = (alpha + beta) + gamma.
Proof.
 intros; generalize plus_assoc3; unfold f_alpha_beta, g_alpha_beta.
 intros;symmetry;auto.
Qed.

End proof_of_associativity.
 
Theorem plus_assoc : forall alpha beta gamma,
                       ordinal alpha ->
                       ordinal beta ->
                       ordinal gamma ->
                       alpha + (beta + gamma) = (alpha + beta) + gamma.
 intros; apply plus_assoc'; auto.
Qed.

Lemma one_plus_ge_omega : forall alpha, omega <= alpha ->
                          F 1 + alpha = alpha. 
Proof.
 intros alpha H;  assert (H0 :ordinal alpha) by (eauto with schutte).
 generalize (minus_exists H); intros [gamma [Og e]];  subst alpha ;
  auto with schutte.
 rewrite plus_assoc;auto with schutte; now rewrite one_plus_omega.
Qed.

Lemma finite_plus_ge_omega : forall n alpha, omega <= alpha ->
                                F n + alpha = alpha.
Proof.
 induction n.
 - simpl; intros;rewrite zero_plus_alpha;trivial; eauto with schutte.
 -  intros; simpl; replace (succ (F n)) with (F 1 + F n).
    + rewrite <- plus_assoc; eauto with schutte.
      rewrite IHn;auto.
      apply one_plus_ge_omega;eauto with schutte.
  + now   rewrite <- plus_FF.
Qed.

Lemma plus_mono_weak_l : forall alpha beta gamma,
                          ordinal gamma ->
                          alpha <= beta -> alpha + gamma <= beta + gamma.
Proof with  eauto with schutte.
 intros alpha beta gamma H H0;
 assert (H1 : ordinal alpha) by ( eauto with schutte).
 assert (H2 :ordinal beta) by ( eauto with schutte).
 case (minus_exists H0); intros  khi [Okhi ekhi];  subst beta.
 rewrite <- plus_assoc ... 
 apply plus_mono_r_weak ... 
 apply le_plus_r ...
Qed.

Lemma plus_mono_bi : forall alpha beta gamma delta, 
                        alpha <= gamma ->
                        beta < delta -> 
                        alpha + beta < gamma + delta.
Proof with eauto with schutte.
 intros alpha beta gamma delta H H0; apply le_lt_trans with (gamma+beta).
 apply plus_mono_weak_l ...
 apply plus_mono_r ...
Qed.

Lemma ordinal_mult_Sn : forall n alpha, ordinal alpha ->
                                         ordinal (mult_Sn alpha n).
Proof.  induction n;simpl;auto with schutte.   Qed.

Hint Resolve ordinal_mult_Sn : schutte.
 
 Lemma mult_n_one : forall n, mult_Sn (F 1) n = F (S n).
 Proof.
  induction n.
  -   trivial.
  -   simpl in *; rewrite IHn; rewrite plus_of_succ; auto with schutte.   
      rewrite alpha_plus_zero;auto with schutte.
 Qed.

Lemma mult_n_mono : forall alpha beta , alpha < beta -> 
   forall n,  mult_Sn alpha n < mult_Sn beta n.
Proof with auto.
 induction n; simpl ...
 apply plus_mono_bi ...
 right ...
Qed.
 
Lemma le_a_mult_Sn_a : forall alpha n, ordinal alpha -> 
                                       alpha <= mult_Sn alpha n.                           
Proof.
 intros alpha n H;  case n.
 -  simpl;auto with schutte.
 -  intros;simpl;apply le_plus_r;auto with schutte.
Qed.

Lemma mult_Sn_mono2 : forall a, zero < a ->
                         forall n p, (n < p)%nat -> mult_Sn a n < mult_Sn a p.
Proof with auto with schutte.
 intros a Ha; assert (ordinal a);[ordi|induction 1 ].
 - simpl; pattern ( mult_Sn a n) at 1; rewrite <- alpha_plus_zero ...
  +   apply plus_mono_r ...
 -  simpl;apply lt_trans with (mult_Sn a m);auto.
    pattern ( mult_Sn a m) at 1; rewrite <- alpha_plus_zero ...
    apply plus_mono_r ...
Qed.

Lemma mult_Sn_mono3 : forall alpha, zero < alpha ->
                         forall n p, (n < p)%nat -> mult_Sn alpha n + alpha 
                                                    <= mult_Sn alpha p.
Proof with auto with schutte.
 intros a Ha; assert (ordinal a);[ordi|induction 1 ].
 - simpl; apply eq_le ...
 - simpl; apply plus_mono_weak_l ...
   apply lt_le ...
   apply  mult_Sn_mono2 ...
Qed.



  
 
  
