(** Well founded relations *)


Require Import Relations Wf_nat List.
Set Implicit Arguments.

Section Proof_of_inverse_image.
Variables (A B:Type) (RA : relation A)(RB : relation B).

Infix "<A" := RA (at level 70, no associativity).
Infix "<B" := RB (at level 70, no associativity).
Variable f : A -> B.

Hypothesis f_mono : forall a a',  a <A a' -> f a <B f a'.
Hypothesis Wf_B : well_founded RB.

Theorem Inverse_image : well_founded RA.
Proof.
  red.
  (*
  1 subgoal, subgoal 1 (ID 6)
  
  A, B : Type
  RA : relation A
  RB : relation B
  f : A -> B
  f_mono : forall a a' : A, a <A a' -> f a <B f a'
  Wf_B : well_founded RB
  ============================
   forall a : A, Acc RA a
  *)

  (* On cherche à faire une récurrence sur B , on  
    introduit alors B dans le but *) 

  pose (P := fun (b : B) =>  forall a:A, f a = b -> Acc RA a).
  assert (forall b:B, P b).
  {  apply (well_founded_induction Wf_B);  unfold P ; intros.
    
  (* 
   f_mono : forall a a' : A, a <A a' -> f a <B f a'
  Wf_B : well_founded RB
  P := fun b : B => forall a : A, f a = b -> Acc RA a : B -> Prop
  x : B
  H : forall y : B, y <B x -> forall a : A, f a = y -> Acc RA a
  a : A
  H0 : f a = x
  ============================
   Acc RA a
  *)
     constructor; intros a' Ha'.
    (*
  f_mono : forall a a' : A, a <A a' -> f a <B f a'
  Wf_B : well_founded RB
  x : B
  H : forall y : B, y <B x -> forall a : A, f a = y -> Acc RA a
  a : A
  H0 : f a = x
  a' : A
  Ha' : a' <A a
  ============================
   Acc RA a'
     *)
   subst x; apply (H (f a')); auto.
   }
    (*
   P := fun b : B => forall a : A, f a = b -> Acc RA a : B -> Prop
  H : forall b : B, P b
  ============================
   forall a : A, Acc RA a
   *) 
  
   intro a ;now apply (H (f a)).
   
Qed.

End Proof_of_inverse_image.


Arguments Inverse_image {A B RA}  RB f _ _ _.

Definition shorter (l l' : list nat) := length l < length l'.

Lemma Wf_shorter : forall A: Type, well_founded shorter.
Proof.
  intro A. apply (Inverse_image   lt  (@length nat)); auto.
  apply lt_wf.
Qed.


Section Pas_descendre.
  Variables (A:Type) (RA: relation A).
  Hypothesis WfA : well_founded RA.

  Infix "<" := RA.
  
  Section Descente.
    Variable suite : nat -> A.
    Hypothesis decr : forall i:nat, suite (S i) < suite  i.

    Let P (a:A) := forall i : nat, a = suite i -> ~ Acc RA a.

    Fact F : forall a:A, P a.
    Proof.
      intro a; apply (well_founded_induction WfA).
      intros x H i Hi H0.
      destruct H0.  specialize (H (suite (S i))).
      subst x;      destruct (H (decr i) (S i)); auto.
    Qed.

    Theorem not_decr : False.
    Proof.      
      now destruct (F (a:=suite 0) (i:=0) ).
    Qed.

  End Descente. 

End Pas_descendre.

